/****************************************************************************
   Program:     $Id: spoofer.c 540 2009-03-27 17:05:09Z rbeverly $
   Author:      Rob Beverly <rbeverly at csail.mit.edu>
   Date:        $Date: 2009-03-27 13:05:09 -0400 (Fri, 27 Mar 2009) $
   Description: Spoofer - generate spoofed packets
****************************************************************************/
#include "spoof.h"

enum debugLevel {OFF, LOW, HIGH} verbose = OFF;

void usage(char *prog) {
    fatal("Usage %s [-hf] [-v]\n", prog);
}

void procErr(int sock) {
    unsigned char buf[BIGBUF];
    spoof_msg_t *msg = (spoof_msg_t *) buf;
    int n;

    memset(buf, 0, BIGBUF);
    n = errorRequest(buf);
    Writen(sock, buf, n);
    n = readMsg(sock, &msg, ERR);
    if ( (n > 0) && (msg->ack == 1) && (msg->flag == 1) ) 
        warn("%s", &buf[sizeof(spoof_msg_t)]);
    else
        warn("Error, but unable to ask server why.");
}

void sendProbes(int sock, int num, test_t *test_ptr, 
               struct in_addr clientip, unsigned int sessionid, int spoof) {
    int i,j;
    unsigned char buf[BIGBUF];
    unsigned char pkt[BIGBUF];
    struct sockaddr_in src, dst;
    int packetlen = 0;
    int displayCounter = 1;
    unsigned int netSessId = 0;
    int ret;

    src.sin_family = AF_INET;
    src.sin_port = htons(SRC_PORT);
    dst.sin_family = AF_INET;
    dst.sin_port = htons(SERV_PORT);
    netSessId = htonl(sessionid);

    memset(buf, 0, BIGBUF);
    memset(pkt, 0, BIGBUF);
    if (verbose < HIGH) {
        if (spoof == TRUE) {
            printf(">> Sourcing spoofed probes (please wait): ");
        } else {
            printf(">> Sourcing non-spoofed validation probes: ");
        }
    }

    for (i=0; i<num; i++, test_ptr++) {
        dst.sin_addr.s_addr = test_ptr->dst_addr;

        /* craft probe packets */
        if ( (spoof == TRUE) && (test_ptr->src_addr != clientip.s_addr) ) {
            src.sin_addr.s_addr = test_ptr->src_addr;
            craftReport(buf, 0, test_ptr);
            packetlen = craftPacket(pkt, buf, PROBESIZE, &src, &dst, TTL_OUT, 0);
            if (verbose >= HIGH) {
                printf(">> Source %d spoofed packets (%s->", DUPLICATES,  inet_ntoa(src.sin_addr));
                printf("%s)...\n", inet_ntoa(dst.sin_addr));
            } else {
                printf("%d..", displayCounter++);
            }
        } else if ( (spoof == FALSE) && (test_ptr->src_addr == clientip.s_addr) ) {
            craftReport(pkt, 1, test_ptr);
            if (verbose >= HIGH) {
                printf(">> Source %d non-spoofed packets to %s...\n", DUPLICATES, inet_ntoa(dst.sin_addr));
            } else {
                printf("..");
            }
        } else {
            continue;
        }

        /* send probe packets */
        for (j = 0; j < DUPLICATES; j++) {
            if (spoof) {
                ret = socketSendSpoof(sock, pkt, packetlen, &dst, sizeof(dst));
                if (ret == WINRAWSOCK_ERROR) {
                    /* after winrawsock err, socketSendSpoof uses raw 
                    ethernet frames.  try again. */
                    j--;
                } else if (ret == SOCKET_ERROR) {
                    strncpy((char *)test_ptr->seqno, "SENDTOFAIL", strlen("SENDTOFAIL")+1);
                    /* don't try to send duplicates on error */
                    j += DUPLICATES;
                } else if (ret == SENDTO_ERROR) {
                    strncpy((char *)test_ptr->seqno, "SENDTOFAIL", strlen("SENDTOFAIL")+1);
                }
            } else {
                memcpy(pkt + PROBESIZE, &netSessId, sizeof(sessionid));
                socketSend(sock, pkt, PROBESIZE + sizeof(netSessId), &dst, sizeof(dst));
            }
            spoofSleep();
        }
    }
    printf("\n");
}

void sendTraceFilter(int sock, uint32_t srcaddr, int pathdist, unsigned int sessionid) {
    unsigned char pkt[BIGBUF];
    struct sockaddr_in src, dst;
    unsigned char *icmpkey;
    int packetlen;
    int i,j;
    unsigned short sid_hi = 0, sid_lo = 0;

    src.sin_family = AF_INET;
    dst.sin_family = AF_INET;
    dst.sin_port = htons(SERV_PORT);
    src.sin_addr.s_addr = srcaddr;
    memset(pkt, 0, BIGBUF);
    /* pick a destination address congruent to source */
    if (ntohl(srcaddr) & 1)
        dst.sin_addr.s_addr = htonl(ntohl(srcaddr) - 1);
    else
        dst.sin_addr.s_addr = htonl(ntohl(srcaddr) + 1);

    /* we encode the low 2B of session ID into the UDP source port 
     * and the high 2B into the IPID field */
    sid_lo = sessionid & 0x0000ffff;
    sid_hi  = ((sessionid & 0xffff0000) >> 16) & 0x0000ffff;
    src.sin_port = htons(sid_lo);
    for (i = 1; i <= pathdist; i++) {
        /* the TTL of the packet will be encoded as the size of the spoofed
         * UDP datagram.  TTL = UDPlength - 8 */
        if ((icmpkey = calloc(i, sizeof(char))) == NULL)
            fatal("calloc error in %s\n", __func__);
        packetlen = craftPacket(pkt, icmpkey, i, &src, &dst, i, sid_hi);
        printf("%d..", i);
        for (j = 0; j < DUPLICATES; j++) {
            socketSendSpoof(sock, pkt, packetlen, &dst, sizeof(dst));
            spoofSleep();
        }
        free(icmpkey);
    }
}

void runTraceFilters(int ctrlsock, int spoofsock, unsigned int sessionid) {
    unsigned char buf[BIGBUF];
    spoof_msg_t *msg = (spoof_msg_t *) buf;
    addr_list_t *list = (addr_list_t *) (buf + sizeof(spoof_msg_t));
    tracefilter_test_t *test = (tracefilter_test_t *) (buf + sizeof(spoof_msg_t) + sizeof(addr_list_t));
    int n;
    int i;

    memset(buf, 0, BIGBUF);
    n = traceFilterRequest(buf);
    Writen(ctrlsock, buf, n);
    readMsg(ctrlsock, &msg, TRACEFILT);
    printf(">> Performing tracefilter [TTL: ");
    if (ntohs(list->size) > 100) {
        warn("Bogus test list.");
        return;
    }
    for (i=0;i<ntohs(list->size);i++) {
        /* Sanity check path distance */
        if (test->dist > TTL_OUT)
            test->dist = 5;
        sendTraceFilter(spoofsock, test->dst_ip, test->dist, sessionid);
        test++;
    }
    printf("]\n");
}

void collectResults(int sock, int num, test_t *test_ptr, 
        struct in_addr clientip, int nat) {
    unsigned char buf[BIGBUF];
    int i, n;
    spoof_msg_t *msg = (spoof_msg_t *) buf;
    
    memset(buf, 0, BIGBUF);
    printf(">> Collecting Results: ");
    if (verbose >= LOW) printf("\n");

    for (i = 0; i < num; i++) {
        n = reportRequest(buf, test_ptr->seqno, test_ptr->src_addr, test_ptr->dst_addr, nat);
        Writen(sock, buf, n);
        if (readMsg(sock, &msg, REPORT) > 0) {
            if (verbose >= LOW) {
                printTest(test_ptr, VERBOSE);
                printf("Received: %d\n",  msg->ack);
            } else {
	            printf("%d..", i);
            }
        }
        test_ptr++;
    }
    printf("\n");
}

void runTraces(int sock, int num, test_t *test_ptr, struct in_addr clientip) {
    unsigned char buf[BIGBUF];
    spoof_msg_t *msg = (spoof_msg_t *) buf;
    char *tracebuf;
    char *bufptr;
    struct in_addr *dests;
    test_t *test_ptr_head = test_ptr;
    int numDst = 0;
    int testsRemain = 0, testsRun = 0, runSize = 0;
    int batchSize = MAX_TRACEROUTE_CHILD;
    int i;

    memset(buf, 0, BIGBUF);
    /* first pass, find destinations to which we want to trace */
    for (i=0; i<num; i++, test_ptr++) {
        if (test_ptr->src_addr == clientip.s_addr) {
            numDst++;
        } 
    }

    /* create destination list */
    if ((dests = calloc(numDst, sizeof(struct in_addr))) == NULL)
        fatal("calloc error in %s\n", __func__);
    test_ptr = test_ptr_head;
    for (i=0; i<numDst; test_ptr++) {
        if (test_ptr->src_addr == clientip.s_addr) {
            dests[i++].s_addr = test_ptr->dst_addr; 
        }
    }

    /* do work */
    testsRemain = numDst;
    while (testsRemain > 0) {
        if (testsRemain - batchSize >= 0) 
            runSize = batchSize;
        else
            runSize = testsRemain;
        testsRemain-=runSize;
        printf(">> Running Traces (please wait) [%d/%d]...", 
            testsRun, numDst);
        if ((tracebuf = calloc(runSize, BIGBUF)) == NULL)
            fatal("calloc error in %s\n", __func__);
        if (runtrace(&dests[testsRun], tracebuf, runSize) < 0) 
            warn("No traceroute found on system");
        printf("\n");
        /* send trace batch */
        for (i=0;i<runSize;i++) {
            bufptr = tracebuf + (BIGBUF*i);
            defaultMsg(msg);
            msg->type = TRACE;
            msg->len = htons(strlen(bufptr) + sizeof(struct in_addr));
            /* place 4B trace destination address ahead of trace string */
            memcpy(buf + sizeof(spoof_msg_t), &dests[testsRun+i], sizeof(struct in_addr));
            /* copy the correct trace string into place */
            memcpy(buf + sizeof(spoof_msg_t) + sizeof(struct in_addr), bufptr, strlen(bufptr));
            Writen(sock, buf, sizeof(spoof_msg_t) + sizeof(struct in_addr) + strlen(bufptr));
            if (readMsg(sock, &msg, TRACE) <= 0) {
                printf(">> Bad server TRACE response\n");
                continue;
            }
            if (msg->ack == 0) {
                printf(">> Bad server TRACE acknowledgement\n");
                procErr(sock);
                continue;
            }
        }
        testsRun += runSize;
        free(tracebuf);
    }
    
    /* clean up */
    free(dests);
}

int checkNAT(int sock) {
    struct sockaddr_in src;
    int srclen;

    memset(&src, 0, sizeof(struct sockaddr_in));

    srclen = sizeof(src);
    if (getsockname(sock, (struct sockaddr *) &src, &srclen) < 0) {
        warn("Couldn't determine my outgoing interface's IP!");
        return TRUE;
    } 
    if (verbose >= HIGH) 
        printf(">> Using outgoing interface IP: %s\n", inet_ntoa(src.sin_addr));
    if (checkRFC1918(&src.sin_addr)) {
        warn("Client appears to be behind a NAT.");
        return TRUE;
    }
    return FALSE;
}

int grabTestScenario(int ctrlsock, test_t **testlistptr) {
    unsigned char buf[BIGBUF];
    spoof_msg_t *msg = (spoof_msg_t *) buf;
    addr_list_t *addr = (addr_list_t *) (buf + sizeof(spoof_msg_t));
    test_t *testlist = NULL;
    int numProbes;
    int i,n;

    n = probeRequest(buf);
    Writen(ctrlsock, buf, n);
    n = readMsg(ctrlsock, &msg, PROBE_REQ);
    if ( (n < 0) || (msg->ack != TRUE) || (msg->flag != TRUE)) {
        procErr(ctrlsock);
        fatal("Couldn't grab test scenario from server.\n");
    } 
    numProbes = ntohs(addr->size);
    if (numProbes <= 0) {
        procErr(ctrlsock);
        fatal("Couldn't grab test scenario from server.\n");
    }
    if (numProbes > 400) {
        procErr(ctrlsock);
        fatal("Bogus test list from server.\n");
    }
    if ((*testlistptr = calloc(numProbes, sizeof(test_t))) == NULL)
        fatal("calloc error in %s\n", __func__);
    testlist = *testlistptr;
    for (i=0;i<numProbes;i++) {
        memcpy(testlist + i, 
            buf + sizeof(spoof_msg_t) + sizeof(addr_list_t) + (sizeof(test_t)*i), 
            sizeof(test_t));
    }
    printf(">> Test scenario with %d probes\n>>\n", numProbes);
    return (numProbes);
}

void sendDone(int ctrlsock, char *sessionkey) {    
    unsigned char buf[BIGBUF];
    spoof_msg_t *msg = (spoof_msg_t *) buf;
    int n;

    printf(">> Finishing test...\n");
    n = doneRequest(buf);
    Writen(ctrlsock, buf, n);
    if (readMsg(ctrlsock, &msg, DONE) <= 0) {
        printf(">> Bad server DONE response\n");
        snprintf(sessionkey, SMALLBUF, "ERR");
    } else {
        snprintf(sessionkey, ntohs(msg->len) + 1, "%s", buf + sizeof(spoof_msg_t));
    }
    if (msg->ack == 0) {
        printf(">> Bad server DONE acknowledgement\n");
        procErr(ctrlsock);
        snprintf(sessionkey, SMALLBUF, "ERR");
    }
}

int sendHello(int ctrlsock, struct in_addr *clientip, unsigned int *sessionid) {    
    unsigned char buf[BIGBUF];
    spoof_msg_t *msg = (spoof_msg_t *) buf;
    hello_msg_t *hello = (hello_msg_t *) (buf + sizeof(spoof_msg_t));
    int n;

    memset(buf, 0, BIGBUF);
    printf(">> Negotiating test scenario with server...\n");
    n = helloRequest(buf);
    Writen(ctrlsock, buf, n);
    if (readMsg(ctrlsock, &msg, HELLO) < sizeof(spoof_msg_t))
        fatal("Connected, but unable to communicate with server (proxy interception?)!\n");    
    
    /* Ensure sane server response */
    if (msg->type != HELLO)
        fatal("Connected, but unable to communicate with server!\n");    
#ifndef REGRESS
    if (msg->ver != VERSION) {
        procErr(ctrlsock);
        warn("Please upgrade your client software\n");
        return -1;
    }
#endif

    /* Play nicely. */
    if (msg->ack == 0) {
        procErr(ctrlsock);
        fatal("Client unable to run test.  Bye.\n");
    }

    /* Valid hello */
    clientip->s_addr = hello->ip;
    *sessionid = ntohl(hello->sessionid);
    return 1;
}

void displayResults(char *sessionkey) {
    unsigned char buf[BUFSIZE];
    snprintf((char *)buf, BUFSIZE, "http://%s/report.php?sessionkey=%s", REPORT_HOST, sessionkey);
    printf("\nTest Complete.\n");
    printf("Your test results:\n\t%s\n\n", buf);
#ifdef _WIN32
    deregisterEther();
    /* spawn windows browser window with the results */
    ShellExecute(NULL, "open", buf, NULL, NULL, SW_SHOWNORMAL);
    winpause("Press Enter to Exit.");
#endif
#ifdef _OSX
    /* spawn osx browser window with the results */
    snprintf((char *)buf, BUFSIZE, "open http://%s/report.php?sessionkey=%s", REPORT_HOST, sessionkey);
    system((char *)buf);
#endif
}

int serverConnect(int *sock, struct sockaddr_in *dst) {
    memset(dst, 0, sizeof(struct sockaddr_in));
    dst->sin_family = AF_INET;
    dst->sin_addr.s_addr = inet_addr(SERVER);
    dst->sin_port = htons(REPORT_PORT);
    *sock = newSocket(SOCK_STREAM, IPPROTO_TCP);
    printf(">> Connecting to server...\n");
    connectSock(*sock, dst, sizeof(*dst));
    return 1;
}

int main(int argc, char **argv) {
    struct sockaddr_in dst;
    struct in_addr clientip;
    test_t *tests = NULL;
    char sessionkey[SMALLBUF];
    unsigned int sessionid = 0;
    int spoofsock = 0, ctrlsock = 0, udpsock = 0;
    int numProbes = 0;
    int noConfirm = FALSE;
    int nat = FALSE;
    int n;

    banner();

    while ((n = getopt(argc, argv, "fhv")) != EOF) {
        switch((char) n) {
        case 'f':
            noConfirm++;
            break;
        case 'v':
            verbose++;
            break;
        case 'h':
        default:
            usage(argv[0]);
        }
    }

    /* Init */
    streamSetup();
    checkroot();
    banner2();
    srand(time(NULL));

    /* Ask for confirmation before proceeding */
    printf(">> Ready to test (please allow several minutes to complete).\n");
    if (!noConfirm)
        confirm(); 

#ifdef _WIN32
    win_init();
    if (findInterfaces() > 1) {
        printf("** Warning: found more than one numbered network interface:\n");
        selectInterface();
    }
#endif

#ifdef REGRESS 
  #include "regress.c"
#endif 

    /* Establish TCP control connection */
    serverConnect(&ctrlsock, &dst);

    /* Check if outgoing interface appears to be behind a NAT */
    nat = checkNAT(ctrlsock);

    /* Create non-spoofing UDP socket for sanity check */
    udpsock = newSocket(SOCK_DGRAM, IPPROTO_UDP);

    /* Force TTL for non-spoofed UDP packets (measure hop distance) */
    optSocketTTL(udpsock, TTL_OUT);

    /* Create UDP spoofing socket */
    spoofsock = newSocket(SOCK_RAW, IPPROTO_UDP);
    optSocketHDRINCL(spoofsock);

    /* Say Hello. */
    if ((sendHello(ctrlsock, &clientip, &sessionid) < 0)) {
        snprintf(sessionkey, SMALLBUF, "BADVERSION");
        goto finalize;
    }

    /* Grab test list */
    numProbes = grabTestScenario(ctrlsock, &tests);

    /* Make certain non-spoofed packets can reach the server */
    sendProbes(udpsock, numProbes, tests, clientip, sessionid, FALSE);

#ifdef _WIN32
    /* wait to register ethernet until we've already communicated with
       the server -- this ensures there is an ARP entry waiting for us */
    registerEther();
#endif

    /* Send spoofed packets */
    sendProbes(spoofsock, numProbes, tests, clientip, 0, TRUE);

    /* Send TTL-limited reverse traceroute spoofed packets */
    runTraceFilters(ctrlsock, spoofsock, sessionid);

    /* We send this final round of non-spoofed UDP packets to determine 
       if the TTL (hop-count) changed during the test */
    sendProbes(udpsock, numProbes, tests, clientip, sessionid, FALSE);

    /* Confer with server to get results */
    collectResults(ctrlsock, numProbes, tests, clientip, nat);

    /* Run traceroutes to each DUT */
    runTraces(ctrlsock, numProbes, tests, clientip);
    
    /* Finish */
    sendDone(ctrlsock, sessionkey);
    
finalize:
    close(udpsock);
    close(spoofsock);
    close(ctrlsock);
    if (tests)
    	free(tests);
    displayResults(sessionkey);
    exit(0);
}
